/*
 *    GS2MGPPSearch.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package org.greenstone.gsdl3.service;

// greenstone classes
import org.greenstone.mgpp.*;
import org.greenstone.gdbm.*;
import org.greenstone.gsdl3.util.*;
// xml classes
import org.w3c.dom.Document; 
import org.w3c.dom.Node; 
import org.w3c.dom.Text; 
import org.w3c.dom.Element; 
import org.w3c.dom.NodeList; 
// general java classes
import java.util.HashMap;
import java.util.Vector;
import java.util.Set;
import java.util.Map;
import java.util.Iterator;
import java.util.Locale;

/**
 * A ServiceRack class for searching in greenstone 2 MGPP collections
 *
 * @author <a href="mailto:kjdon@cs.waikato.ac.nz">Katherine Don</a>
 * @version $Revision: 3680 $
 * @see ServiceRack
 */
public class GS2MGPPSearch 
    extends ServiceRack {

    // these strings must match what is found in the properties file
    // the services on offer
    private static final String TEXT_QUERY_SERVICE = "TextQuery";
    private static final String FIELD_QUERY_SERVICE = "FieldQuery";
    private static final String ADVANCED_FIELD_QUERY_SERVICE = "AdvancedFieldQuery";
    // params used
    private static final String CASE_PARAM = "case";
    private static final String STEM_PARAM = "stem";
    private static final String MATCH_PARAM = "matchMode";
    private static final String MATCH_PARAM_ALL = "all";
    private static final String MATCH_PARAM_SOME = "some";
    private static final String INDEX_PARAM = "index";
    private static final String LEVEL_PARAM = "level";
    private static final String RANK_PARAM = "sortBy";
    private static final String RANK_PARAM_RANK = "rank";
    private static final String RANK_PARAM_NONE = "natural";
    private static final String MAXDOCS_PARAM = "maxDocs";
    private static final String BOOLEAN_PARAM_ON = "1";
    private static final String BOOLEAN_PARAM_OFF = "0";
    private static final String QUERY_PARAM = "query";
    private static final String SIMPLE_FIELD_PARAM = "simpleField";
    private static final String ADVANCED_FIELD_PARAM = "complexField";
    // more params for field query
    private static final String FIELD_QUERY_PARAM = "fqv";
    private static final String FIELD_STEM_PARAM = "fqs";
    private static final String FIELD_CASE_PARAM = "fqc";
    private static final String FIELD_FIELD_PARAM = "fqf";
    private static final String FIELD_COMBINE_PARAM = "fqk";
    private static final String FIELD_COMBINE_PARAM_AND = "0";
    private static final String FIELD_COMBINE_PARAM_OR = "1";
    private static final String FIELD_COMBINE_PARAM_NOT = "2";
    
    // elements used in the config file that are specific to this class
    private static final String DEFAULT_INDEX_ELEM = "defaultIndex";
    private static final String DEFAULT_LEVEL_ELEM = "defaultLevel";
    private static final String INDEX_ELEM = "index";
    private static final String LEVEL_ELEM = "level";
    
    private MGPPWrapper mgpp_src_=null;
    private GDBMWrapper gdbm_src_=null;
    
    private String default_index_=null;
    private String default_level_=null;

    private Element config_info_ = null;
    
    /** constructor */
    public GS2MGPPSearch() {
	mgpp_src_ = new MGPPWrapper();
	gdbm_src_ = new GDBMWrapper();

    }
    
    /** configure this service */
    public boolean configure(Element info) {

	System.out.println("configuring GS2MGPPSearch");
	config_info_ = info;

	// get the default index from <defaultIndex>
	Element def = (Element)GSXML.getChildByTagName(info, DEFAULT_INDEX_ELEM );
	if (def !=null) {
	    default_index_ = def.getAttribute(GSXML.NAME_ATT);
	}
	if (default_index_==null||default_index_.equals("")) {
	    System.err.println("Error: default index not specified!");
	    return false;
	}

	// get the default level out of <defaultLevel>
	def = (Element)GSXML.getChildByTagName(info, DEFAULT_LEVEL_ELEM);
	if (def !=null) {
	    default_level_ = def.getAttribute(GSXML.NAME_ATT);
	}
	if (default_level_==null||default_level_.equals("")) {
	    System.err.println("Error: default level not specified!");
	    return false;
	}
	// the default level is also the level which gdbm is expecting
	mgpp_src_.setReturnLevel(default_level_);

	// set up which services are available for this collection
	Element e = null;
	// these entries should reflect the build config file - some services may not be available depending on how the collection was built.

	// set up short_service_info_ - for now just has name and type
	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_QUERY);
	e.setAttribute(GSXML.NAME_ATT, TEXT_QUERY_SERVICE);
	short_service_info_.appendChild(e);

	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_QUERY);
	e.setAttribute(GSXML.NAME_ATT, FIELD_QUERY_SERVICE);
	short_service_info_.appendChild(e);

	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_QUERY);
	e.setAttribute(GSXML.NAME_ATT, ADVANCED_FIELD_QUERY_SERVICE);
	short_service_info_.appendChild(e);

	// set up service_info_map_ - for now, just has the same elements as above
	// should have full details about each service incl params lists etc.
	// do the text query one - for now a static list. later use buildcfg.xml values to dynamically change this
	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_QUERY);
	e.setAttribute(GSXML.NAME_ATT, TEXT_QUERY_SERVICE);
	Element param_list = doc_.createElement(GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	createTextQueryParamList(param_list, false, null);
	e.appendChild(param_list);
	service_info_map_.put(TEXT_QUERY_SERVICE, e);

	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_QUERY);
	e.setAttribute(GSXML.NAME_ATT, FIELD_QUERY_SERVICE);
	Element param_list1 = doc_.createElement(GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	createFieldQueryParamList(param_list1, false, null);
	e.appendChild(param_list1);
	service_info_map_.put(FIELD_QUERY_SERVICE, e);
	
	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_QUERY);
	e.setAttribute(GSXML.NAME_ATT, ADVANCED_FIELD_QUERY_SERVICE);
	Element param_list2 = doc_.createElement(GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	createAdvancedFieldQueryParamList(param_list2, false, null);
	e.appendChild(param_list2);
	service_info_map_.put(ADVANCED_FIELD_QUERY_SERVICE, e);

	if (gdbm_src_.openDatabase(GSFile.GDBMDatabaseFile(site_home_, cluster_name_), GDBMWrapper.READER)) {
	    return true;
	}
	else {
	    System.err.println("couldn't open gdbm database!");
	    return false;
	}
    }
  
    /** creates a new param element and adds it to the param list */
    protected void createParameter(String name, Element param_list, boolean display, String lang) {
	Element param=null;
	
	if (name.equals(INDEX_PARAM)) {
	    // the index info - read from config file
	    Element index_list = (Element)GSXML.getChildByTagName(config_info_, INDEX_ELEM+GSXML.LIST_MODIFIER);
	    NodeList indexes = index_list.getElementsByTagName(INDEX_ELEM);
	    int len = indexes.getLength();
	    if (len > 1) { // add index param to list only if more than one index specified
		String [] inds  = new String[len];
		for (int i=0; i<len; i++) {
		    inds[i] = ((Element)indexes.item(i)).getAttribute(GSXML.NAME_ATT);
		}
		if (display) {
		    // use the same index names for now - should get these out of the config info
		    param = GSXML.createParameterDisplay(doc_, INDEX_PARAM, getTextString("param."+INDEX_PARAM, lang), inds, inds);
		} else {
		    param = GSXML.createParameter(doc_, INDEX_PARAM, GSXML.PARAM_TYPE_ENUM_SINGLE, default_index_, inds);
		}
	    } 
	} else if (name.equals(LEVEL_PARAM)) {
	    // the level info - read from config file
	    Element level_list = (Element)GSXML.getChildByTagName(config_info_, LEVEL_ELEM+GSXML.LIST_MODIFIER);
	    NodeList levels = level_list.getElementsByTagName(LEVEL_ELEM);
	    int len = levels.getLength();
	    if (len > 1) { // add level param to list only if more than one index specified
		String [] levs  = new String[len];
		for (int i=0; i<len; i++) {
		    levs[i] = ((Element)levels.item(i)).getAttribute(GSXML.NAME_ATT);
		}
		if (display) {
		    param =  GSXML.createParameterDisplay(doc_, LEVEL_PARAM, getTextString("param."+LEVEL_PARAM, lang), levs, levs);
		} else {
		    param = GSXML.createParameter(doc_, LEVEL_PARAM, GSXML.PARAM_TYPE_ENUM_SINGLE, default_level_, levs);
		}
	    }
	} else if (name.equals(STEM_PARAM)) {
	    if (display) {
		String[] bool_ops = {"0", "1"};
		String[] bool_texts = {getTextString("param.boolean.off", lang),getTextString("param.boolean.on", lang)}; 
		param = GSXML.createParameterDisplay(doc_, STEM_PARAM, getTextString("param."+STEM_PARAM, lang),  bool_ops, bool_texts);
	    } else {
		param = GSXML.createParameter(doc_, STEM_PARAM, GSXML.PARAM_TYPE_BOOLEAN, BOOLEAN_PARAM_ON, null);
	    }
	} else if (name.equals(CASE_PARAM)) {
	    if (display) {
		String[] bool_ops = {"0", "1"};
		String[] bool_texts = {getTextString("param.boolean.off", lang),getTextString("param.boolean.on", lang)}; 
		param = GSXML.createParameterDisplay(doc_, CASE_PARAM, getTextString("param."+CASE_PARAM, lang),  bool_ops, bool_texts);
	    } else {
		param = GSXML.createParameter(doc_, CASE_PARAM, GSXML.PARAM_TYPE_BOOLEAN, BOOLEAN_PARAM_ON, null);
	    }
	} else if (name.equals(MATCH_PARAM)) {
	    String []vals = {MATCH_PARAM_ALL, MATCH_PARAM_SOME};
	    if (display) {
		String []val_texts = {getTextString("param."+MATCH_PARAM+"."+MATCH_PARAM_ALL, lang),getTextString("param."+MATCH_PARAM+"."+MATCH_PARAM_SOME, lang)}; 

		param = GSXML.createParameterDisplay(doc_, MATCH_PARAM, getTextString("param."+MATCH_PARAM, lang), vals, val_texts);
	    } else {
		param = GSXML.createParameter(doc_, MATCH_PARAM, GSXML.PARAM_TYPE_ENUM_SINGLE, MATCH_PARAM_ALL, vals);
	    }	
	} else if (name.equals(RANK_PARAM)) {
	    String [] vals1 = {RANK_PARAM_RANK, RANK_PARAM_NONE };
	    if (display) {
		String [] vals1_texts = { getTextString("param."+RANK_PARAM+"."+RANK_PARAM_RANK, lang),
					  getTextString("param."+RANK_PARAM+"."+RANK_PARAM_NONE, lang)};
		
		param = GSXML.createParameterDisplay(doc_, RANK_PARAM, getTextString("param."+RANK_PARAM, lang), vals1, vals1_texts);
	    } else {
		param = GSXML.createParameter(doc_, RANK_PARAM, GSXML.PARAM_TYPE_ENUM_SINGLE, RANK_PARAM_RANK, vals1 );
	    }
	} else if (name.equals(MAXDOCS_PARAM)) {
	    if (display) {
		param = GSXML.createParameterDisplay(doc_, MAXDOCS_PARAM, getTextString("param."+MAXDOCS_PARAM, lang), null, null);
	    } else {
		param = GSXML.createParameter(doc_, MAXDOCS_PARAM, GSXML.PARAM_TYPE_INTEGER, "10", null);
	    }
	} else if (name.equals(QUERY_PARAM)) {
	    if (display) {
		param = GSXML.createParameterDisplay(doc_, QUERY_PARAM, getTextString("param."+QUERY_PARAM, lang), null, null);
	    } else {
		param = GSXML.createParameter(doc_, QUERY_PARAM, GSXML.PARAM_TYPE_STRING, null, null);
	    }
	} else if (name.equals(FIELD_QUERY_PARAM)) {
	    if (display) {
		param = GSXML.createParameterDisplay(doc_, FIELD_QUERY_PARAM, getTextString("param."+FIELD_QUERY_PARAM, lang), null, null);
	    } else {
		param = GSXML.createParameter(doc_, FIELD_QUERY_PARAM, GSXML.PARAM_TYPE_STRING, null, null);
	    }
	    
	    
	} else if (name.equals(FIELD_CASE_PARAM)) {
	    if (display) {
		String[] bool_ops = {"0", "1"};
		String[] bool_texts = {getTextString("param.boolean.off", lang),getTextString("param.boolean.on", lang)}; 
		param = GSXML.createParameterDisplay(doc_, FIELD_CASE_PARAM, getTextString("param."+FIELD_CASE_PARAM, lang),  bool_ops, bool_texts);
		
	    } else {
		param = GSXML.createParameter(doc_, FIELD_CASE_PARAM, GSXML.PARAM_TYPE_BOOLEAN, BOOLEAN_PARAM_ON, null);
		
	    }
	} else if (name.equals(FIELD_STEM_PARAM)) {
	    if (display) {
		String[] bool_ops = {"0", "1"};
		String[] bool_texts = {getTextString("param.boolean.off", lang),getTextString("param.boolean.on", lang)}; 
		param = GSXML.createParameterDisplay(doc_, FIELD_STEM_PARAM, getTextString("param."+FIELD_STEM_PARAM, lang),  bool_ops, bool_texts);
	    } else {
		param = GSXML.createParameter(doc_, FIELD_STEM_PARAM, GSXML.PARAM_TYPE_BOOLEAN, BOOLEAN_PARAM_ON, null);
		
	    }
	} else if (name.equals(FIELD_FIELD_PARAM)) {
	    
	    // the field list - should be read from config file
	    String [] vals2 = {"TX", "SU", "TI", "OR"};
	    String [] vals2_texts = {"TextOnly", "Subject", "Title", "Organization"};
	    
	    if (display) {
		param = GSXML.createParameterDisplay(doc_, FIELD_FIELD_PARAM, getTextString("param."+FIELD_FIELD_PARAM, lang), vals2, vals2_texts);
		
	    } else {
		param = GSXML.createParameter(doc_, FIELD_FIELD_PARAM, GSXML.PARAM_TYPE_ENUM_SINGLE, "", vals2 );
		
	    }
	} else if (name.equals(FIELD_COMBINE_PARAM)) {
      
	    String []vals = {FIELD_COMBINE_PARAM_AND, FIELD_COMBINE_PARAM_OR, FIELD_COMBINE_PARAM_NOT};
	    if (display) {
		String []val_texts = {getTextString("param."+FIELD_COMBINE_PARAM+"."+FIELD_COMBINE_PARAM_AND, lang), getTextString("param."+FIELD_COMBINE_PARAM+"."+FIELD_COMBINE_PARAM_OR, lang), getTextString("param."+FIELD_COMBINE_PARAM+"."+FIELD_COMBINE_PARAM_NOT, lang)}; 
		
		param = GSXML.createParameterDisplay(doc_, FIELD_COMBINE_PARAM, "", vals, val_texts);
		
	    } else {
		param = GSXML.createParameter(doc_, FIELD_COMBINE_PARAM, GSXML.PARAM_TYPE_ENUM_SINGLE, FIELD_COMBINE_PARAM_AND, vals);
		param.setAttribute(GSXML.PARAM_IGNORE_POS_ATT, "0");
		
	    }	
	}

	// add the param to the list
	if (param !=null) {
	    param_list.appendChild(param);
	}
    }
    /** this creates all teh params and appends them to param_list.
     * if display=true it creates the text strings version
     * otherwise it creates the description version
     */
    protected boolean createTextQueryParamList(Element param_list, boolean display, String lang) {

	// the order they are specified here is the order they appear on 
	// the query form
	createParameter(INDEX_PARAM, param_list, display, lang);
	createParameter(LEVEL_PARAM, param_list, display, lang);
	createParameter(CASE_PARAM, param_list, display, lang);
	createParameter(STEM_PARAM, param_list, display, lang);
	createParameter(MATCH_PARAM, param_list, display, lang);
	createParameter(RANK_PARAM, param_list, display, lang);
	createParameter(MAXDOCS_PARAM, param_list, display, lang);
	createParameter(QUERY_PARAM, param_list, display, lang);
	return true;
    }
    /** this creates all teh params and appends them to param_list.
     * if display=true it creates the text strings version
     * otherwise it creates the description version
     */
    protected boolean createFieldQueryParamList(Element param_list, boolean display, String lang) {

	createParameter(MATCH_PARAM, param_list, display, lang);
	createParameter(LEVEL_PARAM, param_list, display, lang);
	createParameter(CASE_PARAM, param_list, display, lang);
	createParameter(STEM_PARAM, param_list, display, lang);
	createParameter(RANK_PARAM, param_list, display, lang);
	createParameter(MAXDOCS_PARAM, param_list, display, lang);

	// create a multi param for the fields etc
	// text box, field
	Element multiparam = null;
	Element param=null;
	if (display) {
	    param = GSXML.createParameterDisplay(doc_, SIMPLE_FIELD_PARAM, "",  null, null);
	    param_list.appendChild(param);
	} else {
	    multiparam = GSXML.createParameter(doc_, SIMPLE_FIELD_PARAM, GSXML.PARAM_TYPE_MULTI, null, null);
	    multiparam.setAttribute("occurs", "4");
	    param_list.appendChild(multiparam);
	}
	// the components
	if (display) {
	    createParameter(FIELD_QUERY_PARAM, param_list, display, lang);
	    createParameter(FIELD_FIELD_PARAM, param_list, display, lang);
	} else {
	    createParameter(FIELD_QUERY_PARAM, multiparam, display, lang);
	    createParameter(FIELD_FIELD_PARAM, multiparam, display, lang);
	}
	return true;
    }

    protected boolean createAdvancedFieldQueryParamList(Element param_list, boolean display, String lang) {

	// first do index and level params
	createParameter(LEVEL_PARAM, param_list, display, lang);
	createParameter(RANK_PARAM, param_list, display, lang);
	createParameter(MAXDOCS_PARAM, param_list, display, lang);


	// create a multi param for the fields etc
	// text box, stem, case, field

	Element multiparam = null;
	Element param=null;

	if (display) {
	    param = GSXML.createParameterDisplay(doc_, ADVANCED_FIELD_PARAM, "",  null, null);
	    param_list.appendChild(param);
	} else {
	    multiparam = GSXML.createParameter(doc_, ADVANCED_FIELD_PARAM, GSXML.PARAM_TYPE_MULTI, null, null);
	    multiparam.setAttribute("occurs", "4");
	    param_list.appendChild(multiparam);
	}
	
	// the components

	//the combiners
	

	if (display) {
	    createParameter(FIELD_COMBINE_PARAM, param_list, display, lang);
	    createParameter(FIELD_QUERY_PARAM, param_list, display, lang);
	    createParameter(FIELD_CASE_PARAM, param_list, display, lang);
	    createParameter(FIELD_STEM_PARAM, param_list, display, lang);
	    createParameter(FIELD_FIELD_PARAM, param_list, display, lang);
	} else {
	    createParameter(FIELD_COMBINE_PARAM, multiparam, display, lang);
	    createParameter(FIELD_QUERY_PARAM, multiparam, display, lang);
	    createParameter(FIELD_CASE_PARAM, multiparam, display, lang);
	    createParameter(FIELD_STEM_PARAM, multiparam, display, lang);
	    createParameter(FIELD_FIELD_PARAM, multiparam, display, lang);    
	}

	return true;
    }
    /** creates a display element containing all the text strings needed to display the service page, in the language specified */
    protected Element createServiceDisplay(String service, String lang) {
	Element display = doc_.createElement(GSXML.DISPLAY_ELEM);
	display.appendChild(GSXML.createTextElement(doc_, GSXML.DISPLAY_NAME_ELEM,  getTextString(service+".name", lang)));
	display.appendChild(GSXML.createTextElement(doc_,  GSXML.DISPLAY_SUBMIT_ELEM, getTextString(service+".submit", lang)));

	// now need to add in the params
	if (service.equals(TEXT_QUERY_SERVICE)) {
	    
	    createTextQueryParamList(display, true, lang);
	}
	// now need to add in the params
	if (service.equals(FIELD_QUERY_SERVICE)) {
	    
	    createFieldQueryParamList(display, true, lang);
	}
	// now need to add in the params
	if (service.equals(ADVANCED_FIELD_QUERY_SERVICE)) {
	    
	    createAdvancedFieldQueryParamList(display, true, lang);
	}
	return display;
	
    }

    
    /** process a text query */
    protected Element processTextQuery(Element request) {

	Element result = doc_.createElement(GSXML.RESPONSE_ELEM);
	String from = GSPath.appendLink(cluster_name_, TEXT_QUERY_SERVICE);
	result.setAttribute(GSXML.FROM_ATT, from);
	result.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_QUERY);
	
	// get param list
	Element param_list = (Element)GSXML.getChildByTagName(request, GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	if (param_list==null) {
	    System.err.println("GS2MGPPSearch, TextQuery Error: no param list in request!");
	    return result; // empty result
	}
	HashMap params = GSXML.extractParams(param_list);
	String query = (String)params.get(QUERY_PARAM);
	if (query == null) {
	    // nothing to query
	    return result;
	}
	
	String index = (String)params.get(INDEX_PARAM); 
	if (index==null) { // if its not present, use the default index
	    index=default_index_;
	}

	// now set up the mgpp stuff
	String basedir = GSFile.collectionBaseDir(site_home_, 
						  cluster_name_);
	String indexdir = GSFile.collectionIndexPath(cluster_name_, index);

	mgpp_src_.loadIndexData(basedir, indexdir);
	setStandardQueryParams(params);
	
	mgpp_src_.runQuery(query);
	MGPPQueryResult mqr= mgpp_src_.getQueryResult();
	
	// get the docnums out, and convert to HASH ids
	Vector docs = mqr.getDocs();
	if (docs.size()==0) {
	    // no docs found
	    return result;
	}
	// create a documentList element
	Element c = doc_.createElement(GSXML.CONTENT_ELEM);
	result.appendChild(c);
	Element document_list = doc_.createElement(GSXML.DOCUMENT_ELEM+GSXML.LIST_MODIFIER);
	c.appendChild(document_list);
	// add each document
	for (int d=0; d<docs.size(); d++) {
	    long docnum = ((MGPPDocInfo)docs.elementAt(d)).num_;
	    String id = gdbm_src_.docnum2Oid(docnum);
	    Node no = GSXML.createDocumentElement(doc_, id);
	    document_list.appendChild(no);
	}
	return result;
    }//processTextQuery

    protected Element processFieldQuery(Element request) {
	return processAnyFieldQuery(request, "simple");
    }

    protected Element processAdvancedFieldQuery(Element request) {
	return processAnyFieldQuery(request, "advanced");
    }

    /** process a fielded query */
    protected Element processAnyFieldQuery(Element request, String type) {
	
	System.out.println("processing any Field Query...");
	Element result = doc_.createElement(GSXML.RESPONSE_ELEM);
	String from = null;
	if (type.equals("simple")) {
	    from = GSPath.appendLink(cluster_name_, FIELD_QUERY_SERVICE);
	} else {
	    from = GSPath.appendLink(cluster_name_, ADVANCED_FIELD_QUERY_SERVICE);
	}
	result.setAttribute(GSXML.FROM_ATT, from);
	result.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_QUERY);
	
	// get param list
	Element param_list = (Element)GSXML.getChildByTagName(request, GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	if (param_list==null) {
	    System.err.println("GS2MGPPSearch, FieldQuery Error: no param list in request!");
	    return result; // empty result
	}

	HashMap params = GSXML.extractAllParams(param_list);
	printHash(params);

	//check one of the params to see if it is present
	String test_param = (String)params.get(FIELD_QUERY_PARAM);
	if (test_param == null) { // no query
	    return result;
	}

	// for now we are using the default index
	String index = default_index_;
	
	// now set up the mgpp stuff
	String basedir = GSFile.collectionBaseDir(site_home_, 
						  cluster_name_);
	String indexdir = GSFile.collectionIndexPath(cluster_name_, index);
	
	mgpp_src_.loadIndexData(basedir, indexdir);
	setStandardQueryParams(params);
	
	String query="";
	if (type.equals("simple")) {
	    query = parseFieldQueryParams(params);
	} else if (type.equals("advanced")) {
	    query = parseAdvancedFieldQueryParams(params);
	}
	System.out.println("query="+query);
	mgpp_src_.runQuery(query);
	MGPPQueryResult mqr= mgpp_src_.getQueryResult();
	
	// get the docnums out, and convert to HASH ids
	Vector docs = mqr.getDocs();
	if (docs.size()==0) {
	    // no docs found
	    return result;
	}
	// create a documentList element
	Element c = doc_.createElement(GSXML.CONTENT_ELEM);
	result.appendChild(c);
	Element document_list = doc_.createElement(GSXML.DOCUMENT_ELEM+GSXML.LIST_MODIFIER);
	c.appendChild(document_list);
	// add each document
	for (int d=0; d<docs.size(); d++) {
	    long docnum = ((MGPPDocInfo)docs.elementAt(d)).num_;
	    String id = gdbm_src_.docnum2Oid(docnum);
	    Node no = GSXML.createDocumentElement(doc_, id);
	    document_list.appendChild(no);
	}
	return result;
    }



    // should probably use a list rather than map
    protected boolean setStandardQueryParams(HashMap params) {

	// set the default ones
	mgpp_src_.setReturnTerms(true);
	Set entries = params.entrySet();
	Iterator i = entries.iterator();
	while (i.hasNext()) {
	    Map.Entry m = (Map.Entry)i.next();
	    String name = (String)m.getKey();
	    String value = (String)m.getValue();
	   
	    if (name.equals(CASE_PARAM)) {
		boolean val = (value.equals(BOOLEAN_PARAM_ON)?true:false);
		mgpp_src_.setCase(val);		
	    } else if (name.equals(STEM_PARAM)) {
		boolean val = (value.equals(BOOLEAN_PARAM_ON)?true:false);
		mgpp_src_.setStem(val);
	    } else if (name.equals(MAXDOCS_PARAM)) {
		int docs = Integer.parseInt(value);
		mgpp_src_.setMaxDocs(docs);
	    } else if (name.equals(LEVEL_PARAM)) {
		mgpp_src_.setQueryLevel(value);
	    } else if (name.equals(MATCH_PARAM)) {
		int mode;
		if (value.equals(MATCH_PARAM_ALL)) mode=1;
		else mode=0;
		mgpp_src_.setMatchMode(mode);
	    } else if (name.equals(RANK_PARAM)) {
		if (value.equals(RANK_PARAM_RANK)) {
		    mgpp_src_.setSortByRank(true);
		} else if (value.equals(RANK_PARAM_NONE)) {
		    mgpp_src_.setSortByRank(false);
		}
	    } // ignore any others
	}
	return true;
    }

    protected String parseFieldQueryParams(HashMap params) {

	StringBuffer final_query = new StringBuffer(256);
	String text_line = (String)params.get(FIELD_QUERY_PARAM);
	String[] texts = text_line.split(",", -1);
	String field_line = (String)params.get(FIELD_FIELD_PARAM);
	String[] fields = field_line.split(",", -1);
	String combine="&";
	String match = (String)params.get(MATCH_PARAM);
	if (match.equals(MATCH_PARAM_SOME)) {
	    combine = "|";
	}
	
	for (int i=0; i<texts.length; i++) {

	    String q = texts[i].trim();
	    if (!q.equals("")) {
		addQueryElem(final_query, q, fields[i], combine);
	    }
	}

	System.out.println("final query = "+final_query.toString());
	return final_query.toString();
    }
    
    protected String parseAdvancedFieldQueryParams(HashMap params) {

	StringBuffer final_query = new StringBuffer(256);
	String text_line = (String)params.get(FIELD_QUERY_PARAM);
	String[] texts = text_line.split(",", -1);
	String field_line = (String)params.get(FIELD_FIELD_PARAM);
	String[] fields = field_line.split(",", -1);
	String case_line = (String)params.get(FIELD_CASE_PARAM);
	String[] cases = case_line.split(",", -1);
	String stem_line = (String)params.get(FIELD_STEM_PARAM);
	String[] stems = stem_line.split(",", -1);
	String  combine_line = (String)params.get(FIELD_COMBINE_PARAM);
	String [] combines = combine_line.split(",", -1);
	String combine = "&";
	for (int i=0; i<texts.length; i++) {
	    if (i==0) {
		combine = "";
	    } else {
		String x = combines[i-1]; // assume first one is missing
		if (x.equals(FIELD_COMBINE_PARAM_AND)) {
		    combine = "&";
		} else if (x.equals(FIELD_COMBINE_PARAM_OR)) {
		    combine = "|";
		} else if (x.equals(FIELD_COMBINE_PARAM_NOT)) {
		    combine = "!";
		} 
		
	    }

	    String q = texts[i].trim();
	    if (!q.equals("")) {
		q = addStemAndCase(q, stems[i], cases[i]);
		addQueryElem(final_query, q, fields[i], combine);
	    }
	}

	System.out.println("final query = "+final_query.toString());
	return final_query.toString();
    }

    protected void addQueryElem(StringBuffer s, String q, String f, String c) {

	String combine="";
	if (s.length()>0) {
	    combine = " "+c+" ";
	}
	if (f!="") {
	    s.append(combine+"["+q+"]:"+f);
	} else {
	    s.append(combine+q);
	}
    }
    protected String addStemAndCase(String q, String s, String c) {
	String mods = "#";
	if (c.equals("1")) {
	    mods += "i";
	} else {
	    mods += "c";
	}
	if (s.equals("1")) {
	    mods += "s";
	} else {
	    mods+= "u";
	}
	StringBuffer temp = new StringBuffer();
	String [] terms = q.split(" ");
	for (int i=0; i<terms.length; i++) {
	    String t = terms[i].trim();
	    
	    if (!t.equals("") && !t.equals("TX")) {
		temp.append(" "+t+mods);
	    }
	}
	return temp.toString();
    }

    protected void printHash(HashMap map) {
	System.out.println("Map entries are :");
	Set entries = map.entrySet();
	Iterator i = entries.iterator();
	while (i.hasNext()) {
	    Map.Entry m = (Map.Entry)i.next();
	    String name = (String)m.getKey();
	    String value = (String)m.getValue();
	    System.out.println(name+"="+value);
	}  
    }

}
 

