/*
 *    GS2MGRetrieve.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package org.greenstone.gsdl3.service;

// greenstone classes
import org.greenstone.mg.*;
import org.greenstone.gdbm.*;
import org.greenstone.gsdl3.util.*;

// xml classes
import org.w3c.dom.CDATASection;
import org.w3c.dom.Document; 
import org.w3c.dom.Element; 
import org.w3c.dom.Node; 
import org.w3c.dom.NodeList; 
import org.w3c.dom.Text; 

// general Java classes
import java.io.File;
import java.util.HashMap;

/**
 *
 * @author <a href="mailto:kjdon@cs.waikato.ac.nz">Katherine Don</a>
 * @version $Revision: 3754 $
 */

public class GS2MGRetrieve
    extends ServiceRack {

    // the services on offer
    // these strings must match what is found in the properties file
    private static final String DOCUMENT_RETRIEVE_SERVICE = "DocumentRetrieve";
    private static final String METADATA_RETRIEVE_SERVICE = "MetadataRetrieve";

    // params used
    private static final String INDEX_PARAM = "index";

    // elements used in the config file that are specific to this class
    private static final String DEFAULT_INDEX_ELEM = "defaultIndex";

    private MGWrapper mg_src_ = null;
    private GDBMWrapper gdbm_src_ = null;
    
    private String default_index_ = null;


    /** constructor */
    public GS2MGRetrieve() {
	mg_src_ = new MGWrapper();
	gdbm_src_ = new GDBMWrapper();
    }

    /** configure this service */
    public boolean configure(Element info)
    {
	System.out.println("configuring GS2MGRetrieve");

	// get the default index out of <defaultIndex> (buildConfig.xml)
	Element def = (Element) GSXML.getChildByTagName(info, DEFAULT_INDEX_ELEM);
	if (def != null) {
	    default_index_ = def.getAttribute(GSXML.NAME_ATT);
	}
	if (default_index_ == null || default_index_.equals("")) {
	    System.err.println("Error: default index not specified!");
	    return false;
	}
	System.out.println("Default index: " + default_index_);

	Element e = null;
	// these entries should reflect the build config file - some services may not be available depending on how the colleciton was built.
	// set up short_service_info_ - for now just has name and type
	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
	e.setAttribute(GSXML.NAME_ATT, DOCUMENT_RETRIEVE_SERVICE);
	short_service_info_.appendChild(e);

	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
	e.setAttribute(GSXML.NAME_ATT, METADATA_RETRIEVE_SERVICE);
	short_service_info_.appendChild(e);

	// set up service_info_map_ - for now, just has the same elements as above
	// should have full details about each service incl params lists etc.
	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
	e.setAttribute(GSXML.NAME_ATT, DOCUMENT_RETRIEVE_SERVICE);
	service_info_map_.put(DOCUMENT_RETRIEVE_SERVICE, e);

	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
	e.setAttribute(GSXML.NAME_ATT, METADATA_RETRIEVE_SERVICE);
	service_info_map_.put(METADATA_RETRIEVE_SERVICE, e);

	// Open GDBM database for querying
	String gdbm_db_file = GSFile.GDBMDatabaseFile(site_home_, cluster_name_);
	if (gdbm_src_.openDatabase(gdbm_db_file, GDBMWrapper.READER)) {
	    return true;
	}
	else {
	    System.err.println("Error: Could not open gdbm database!");
	    return false;
	}
    }

    /** creates a display element containing all the text strings needed to display the service page, in the language specified 
     * these retrieval services dont get displayed to the users - they are only used internally by the actions. so this returns an empty display element*/
    protected Element createServiceDisplay(String service, String lang)
    {
	return doc_.createElement(GSXML.DISPLAY_ELEM);
    }


    /** retrieve a document */
    protected Element processDocumentRetrieve(Element request)
    {
	// an empty result
	Element result = doc_.createElement(GSXML.RESPONSE_ELEM);
	String from = GSPath.appendLink(cluster_name_, DOCUMENT_RETRIEVE_SERVICE);
	result.setAttribute(GSXML.FROM_ATT, from);
	result.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_QUERY);
	Element result_doc = doc_.createElement(GSXML.CONTENT_ELEM);
	result.appendChild(result_doc);

	// get param list and content 
	Element param_elem = null;
	Element content_elem = null;
	Node n = request.getFirstChild();
	while (n != null) {
	    String node_name = n.getNodeName();
	    if (node_name.equals(GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER)) {
		param_elem = (Element) n;
	    } else if (node_name.equals(GSXML.CONTENT_ELEM)) {
		content_elem = (Element) n;
	    }
	    n = n.getNextSibling();
	}

	if (param_elem==null || content_elem==null) {
	    System.err.println("GS2MGRetrieve: malformed request sent to DocumentRetrieve service");
	    return result; // empty result
	}
	
	HashMap params = GSXML.extractParams(param_elem);

	// get docs from mg
	String index = (String) params.get(INDEX_PARAM); 
	if (index == null) { // if it is not present, use the default index
	    index = default_index_;
	}
	System.out.println("Index: " + index);

	String basedir = GSFile.collectionBaseDir(site_home_, cluster_name_) + File.separatorChar;  // Needed by MG
	String textdir = GSFile.collectionTextPath(cluster_name_);
	String indexpath = GSFile.collectionIndexPath(cluster_name_, index);
	System.out.println("Base directory: " + basedir);
	System.out.println("Text directory: " + textdir);
	System.out.println("Index path: " + indexpath);

	mg_src_.setIndex(indexpath);

	// get the text for each doc in the list
	String[] ids = GSXML.getDocumentNameList(content_elem);
	for (int j=0; j<ids.length; j++) {
	    long real_num = gdbm_src_.oid2Docnum(ids[j]);
	    String document = mg_src_.getDocument(basedir, textdir, real_num);
	    System.out.println("(MGRetrieve) Document: " + document);
	    // for now, stick it in a text node - eventually should be parsed as xml??
	    // something funny with the doc - 
	    Element new_doc = GSXML.createDocumentElement(doc_, ids[j]);
	    GSXML.addDocText(doc_, new_doc, document);

	    result_doc.appendChild(new_doc);
	}
	return result;
    }


    /** retrieve metadata */
    protected Element processMetadataRetrieve(Element request)
    {
	Element result = doc_.createElement(GSXML.RESPONSE_ELEM);
	String from = GSPath.appendLink(cluster_name_, METADATA_RETRIEVE_SERVICE);
	result.setAttribute(GSXML.FROM_ATT, from);
	result.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_QUERY);

	Element result_content = doc_.createElement(GSXML.CONTENT_ELEM);
	result.appendChild(result_content);
	Element document_list = doc_.createElement(GSXML.DOCUMENT_ELEM+GSXML.LIST_MODIFIER);
	result_content.appendChild(document_list);

	// get the metadata
       	Element content = (Element) request.getElementsByTagName(GSXML.CONTENT_ELEM).item(0);
	if (content == null) {
	    // error: query had no content!! - should make an error message
	    return result;
	}
	String[] metas = GSXML.getMetaNameList(content);
	String[] ids = GSXML.getDocumentNameList(content);
	for (int j = 0; j < ids.length; j++) { // for each document
	    Element doc = GSXML.createDocumentElement(doc_, ids[j]);
	    Element list = GSXML.addMetaList(doc_, doc);
	    DBInfo info = gdbm_src_.getInfo(ids[j]);
	    for (int m = 0; m < metas.length; m++) {
		String value = info.getInfo(metas[m]);
		GSXML.addMetadata(doc_, list, metas[m], value);
	    }
	    document_list.appendChild(doc);
	}
	return result;
    }
}   
