/*
 *    GS2MGSearch.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package org.greenstone.gsdl3.service;

// greenstone classes
import org.greenstone.mg.*;
import org.greenstone.gdbm.*;
import org.greenstone.gsdl3.util.*;

// xml classes
import org.w3c.dom.Document; 
import org.w3c.dom.Element; 
import org.w3c.dom.Node;
import org.w3c.dom.NodeList; 
import org.w3c.dom.Text; 

// general java classes
import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

/**
 *
 * @author <a href="mailto:kjdon@cs.waikato.ac.nz">Katherine Don</a>
 * @version $Revision: 3770 $
 */

public class GS2MGSearch 
    extends ServiceRack {

    // the services on offer
    // these strings must match what is found in the properties file
    private static final String TEXT_QUERY_SERVICE = "TextQuery";

    // params used
    private static final String INDEX_PARAM = "index";
    private static final String CASE_PARAM = "case";
    private static final String STEM_PARAM = "stem";
    private static final String MATCH_PARAM = "matchMode";
    private static final String MATCH_PARAM_ALL = "all";
    private static final String MATCH_PARAM_SOME = "some";
    private static final String RANK_PARAM = "sortBy";
    private static final String RANK_PARAM_RANK = "rank";
    private static final String RANK_PARAM_NONE = "natural";
    private static final String MAXDOCS_PARAM = "maxDocs";
    private static final String BOOLEAN_PARAM_ON = "1";
    private static final String BOOLEAN_PARAM_OFF = "0";
    private static final String QUERY_PARAM = "query";

    // elements used in the config file that are specific to this class
    private static final String DEFAULT_INDEX_ELEM = "defaultIndex";
    private static final String INDEX_ELEM = "index";

    public static final String COUNT_ATT = "count";
    public static final String STEM_ATT = "stem";

    private MGWrapper mg_src_ = null;
    private GDBMWrapper gdbm_src_ = null;

    private String default_index_ = null;

    private Element config_info_ = null;


    /** constructor */
    public GS2MGSearch() {
	mg_src_ = new MGWrapper();
	gdbm_src_ = new GDBMWrapper();
    }


    /** configure this service */
    public boolean configure(Element info)
    {
	System.out.println("configuring GS2MGSearch");
	config_info_ = info;

	// get the default index out of <defaultIndex> (buildConfig.xml)
	Element def = (Element) GSXML.getChildByTagName(info, DEFAULT_INDEX_ELEM);
	if (def != null) {
	    default_index_ = def.getAttribute(GSXML.NAME_ATT);
	}
	if (default_index_ == null || default_index_.equals("")) {
	    System.err.println("Error: default index not specified!");
	    return false;
	}

	Element e = null;
	// these entries should reflect the build config file - some services may not be available depending on how the colleciton was built.
	// set up short_service_info_ - for now just has name and type
	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_QUERY);
	e.setAttribute(GSXML.NAME_ATT, TEXT_QUERY_SERVICE);
	short_service_info_.appendChild(e);

	// set up service_info_map_ - for now, just has the same elements as above
	// should have full details about each service incl params lists etc.
	e = doc_.createElement(GSXML.SERVICE_ELEM);
	e.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_QUERY);
	e.setAttribute(GSXML.NAME_ATT, TEXT_QUERY_SERVICE);
	Element param_list = doc_.createElement(GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	createTextQueryParamList(param_list, false, null);
	e.appendChild(param_list);
	service_info_map_.put(TEXT_QUERY_SERVICE, e);

	// Open GDBM database for querying
	String gdbm_db_file = GSFile.GDBMDatabaseFile(site_home_, cluster_name_);
	if (gdbm_src_.openDatabase(gdbm_db_file, GDBMWrapper.READER)) {
	    return true;
	}
	else {
	    System.err.println("Error: Could not open gdbm database!");
	    return false;
	}
    }


    /** creates a new param element and adds it to the param list */
    protected void createParameter(String name, Element param_list, boolean display, String lang)
    {
	Element param=null;

	if (name.equals(INDEX_PARAM)) {
	    // the index info - read from config file
	    Element index_list = (Element)GSXML.getChildByTagName(config_info_, INDEX_ELEM+GSXML.LIST_MODIFIER);
	    NodeList indexes = index_list.getElementsByTagName(INDEX_ELEM);
	    int len = indexes.getLength();
	    if (len > 1) { // add index param to list only if more than one index specified
		String [] inds  = new String[len];
		for (int i=0; i<len; i++) {
		    inds[i] = ((Element)indexes.item(i)).getAttribute(GSXML.NAME_ATT);
		}
		if (display) {
		    // use the same index names for now - should get these out of the config info
		    param = GSXML.createParameterDisplay(doc_, INDEX_PARAM, getTextString("param."+INDEX_PARAM, lang), inds, inds);
		} else {
		    param = GSXML.createParameter(doc_, INDEX_PARAM, GSXML.PARAM_TYPE_ENUM_SINGLE, default_index_, inds);
		}
	    } 
	}
	else if (name.equals(CASE_PARAM)) {
	    if (display) {
		String[] bool_ops = {"0", "1"};
		String[] bool_texts = {getTextString("param.boolean.off", lang),getTextString("param.boolean.on", lang)}; 
		param = GSXML.createParameterDisplay(doc_, CASE_PARAM, getTextString("param."+CASE_PARAM, lang),  bool_ops, bool_texts);
	    } else {
		param = GSXML.createParameter(doc_, CASE_PARAM, GSXML.PARAM_TYPE_BOOLEAN, BOOLEAN_PARAM_ON, null);
	    }
	}
	else if (name.equals(STEM_PARAM)) {
	    if (display) {
		String[] bool_ops = {"0", "1"};
		String[] bool_texts = {getTextString("param.boolean.off", lang),getTextString("param.boolean.on", lang)}; 
		param = GSXML.createParameterDisplay(doc_, STEM_PARAM, getTextString("param."+STEM_PARAM, lang),  bool_ops, bool_texts);
	    } else {
		param = GSXML.createParameter(doc_, STEM_PARAM, GSXML.PARAM_TYPE_BOOLEAN, BOOLEAN_PARAM_ON, null);
	    }
	}
	else if (name.equals(MATCH_PARAM)) {
	    String[] vals = {MATCH_PARAM_ALL, MATCH_PARAM_SOME};
	    if (display) {
		String[] val_texts = {getTextString("param."+MATCH_PARAM+"."+MATCH_PARAM_ALL, lang),getTextString("param."+MATCH_PARAM+"."+MATCH_PARAM_SOME, lang)}; 

		param = GSXML.createParameterDisplay(doc_, MATCH_PARAM, getTextString("param."+MATCH_PARAM, lang), vals, val_texts);
	    } else {
		param = GSXML.createParameter(doc_, MATCH_PARAM, GSXML.PARAM_TYPE_ENUM_SINGLE, MATCH_PARAM_ALL, vals);
	    }	
	}
	else if (name.equals(RANK_PARAM)) {
	    String[] vals1 = {RANK_PARAM_RANK, RANK_PARAM_NONE };
	    if (display) {
		String[] vals1_texts = { getTextString("param."+RANK_PARAM+"."+RANK_PARAM_RANK, lang),
					  getTextString("param."+RANK_PARAM+"."+RANK_PARAM_NONE, lang)};
		
		param = GSXML.createParameterDisplay(doc_, RANK_PARAM, getTextString("param."+RANK_PARAM, lang), vals1, vals1_texts);
	    } else {
		param = GSXML.createParameter(doc_, RANK_PARAM, GSXML.PARAM_TYPE_ENUM_SINGLE, RANK_PARAM_RANK, vals1 );
	    }
	}
	else if (name.equals(MAXDOCS_PARAM)) {
	    if (display) {
		param = GSXML.createParameterDisplay(doc_, MAXDOCS_PARAM, getTextString("param."+MAXDOCS_PARAM, lang), null, null);
	    } else {
		param = GSXML.createParameter(doc_, MAXDOCS_PARAM, GSXML.PARAM_TYPE_INTEGER, "10", null);
	    }
	}
	else if (name.equals(QUERY_PARAM)) {
	    if (display) {
		param = GSXML.createParameterDisplay(doc_, QUERY_PARAM, getTextString("param."+QUERY_PARAM, lang), null, null);
	    } else {
		param = GSXML.createParameter(doc_, QUERY_PARAM, GSXML.PARAM_TYPE_STRING, null, null);
	    }
	}

	// add the param to the list
	if (param != null) {
	    param_list.appendChild(param);
	}
    }


    /** this creates all teh params and appends them to param_list.
     * if display=true it creates the text strings version
     * otherwise it creates the description version
     */
    protected boolean createTextQueryParamList(Element param_list, boolean display, String lang)
    {
	// the order they are specified here is the order they appear on 
	// the query form
	createParameter(INDEX_PARAM, param_list, display, lang);
	createParameter(CASE_PARAM, param_list, display, lang);
	createParameter(STEM_PARAM, param_list, display, lang);
	createParameter(MATCH_PARAM, param_list, display, lang);
	createParameter(RANK_PARAM, param_list, display, lang);
	createParameter(MAXDOCS_PARAM, param_list, display, lang);
	createParameter(QUERY_PARAM, param_list, display, lang);
	return true;
    }


    /** creates a display element containing all the text strings needed to display the service page, in the language specified */
    protected Element createServiceDisplay(String service, String lang) {
	Element display = doc_.createElement(GSXML.DISPLAY_ELEM);
	display.appendChild(GSXML.createTextElement(doc_, GSXML.DISPLAY_NAME_ELEM,
						    getTextString(service+".name", lang)));
	display.appendChild(GSXML.createTextElement(doc_, GSXML.DISPLAY_SUBMIT_ELEM,
						    getTextString(service+".submit", lang)));

	// now need to add in the params
	if (service.equals(TEXT_QUERY_SERVICE)) {
	    createTextQueryParamList(display, true, lang);
	}

	return display;
    }


    /** process a text query */
    protected Element processTextQuery(Element request)
    {
	Element result = doc_.createElement(GSXML.RESPONSE_ELEM);
	String from = GSPath.appendLink(cluster_name_, TEXT_QUERY_SERVICE);
	result.setAttribute(GSXML.FROM_ATT, from);
	result.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_QUERY);
	// result.setAttribute(GSXML.NAME_ATT, TEXT_QUERY_SERVICE);

	// get param list
	Element param_list = (Element)GSXML.getChildByTagName(request, GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	if (param_list==null) {
	    System.err.println("GS2MGSearch, TextQuery Error: no param list in request!");
	    return result; // empty result
	}
	HashMap params = GSXML.extractParams(param_list);
	String query = (String)params.get(QUERY_PARAM);
	if (query == null) {
	    // no query, no result
	    return result;
	}

	String index = (String) params.get(INDEX_PARAM); 
	if (index == null) { // if it is not present, use the default index
	    index = default_index_;
	}
	// System.out.println("Index: " + index);

	// now set up the mg stuff
	String basedir = GSFile.collectionBaseDir(site_home_, cluster_name_) +
	                   File.separatorChar;  // Needed for MG
	String textdir = GSFile.collectionTextPath(cluster_name_);
	String indexpath = GSFile.collectionIndexPath(cluster_name_, index);

	// set the mg query parameters to the values the user has specified
	setStandardQueryParams(params);
	mg_src_.setIndex(indexpath);

	System.out.println("GS2MGSearch, query string: " + query);
	mg_src_.runQuery(basedir + File.separatorChar, textdir, query);
	MGQueryResult mqr = mg_src_.getQueryResult();
	long totalDocs = mqr.getTotalDocs();
	// System.out.println("Matching documents: " + totalDocs);

	// get the docnums out, and convert to HASH ids
	Vector docs = mqr.getDocs();
	if (docs.size() == 0) {
	    // no docs found
	    System.out.println("No results found...\n");
	}

	// Create a metadata list to store information about the query results
	Element metadata_list = GSXML.addMetaList(doc_, result);

	// Add a metadata element specifying the number of matching documents
	Element num_matches_elem = doc_.createElement(GSXML.METADATA_ELEM);
	num_matches_elem.setAttribute(GSXML.NAME_ATT, "numDocsMatched");
	num_matches_elem.setAttribute(GSXML.VALUE_ATT, "" + totalDocs);
	metadata_list.appendChild(num_matches_elem);
	// System.out.println("Metadata list: " + converter_.getString(metadata_list));

	// Response content: documents and terms
	Element c = doc_.createElement(GSXML.CONTENT_ELEM);
	result.appendChild(c);

	Element document_list = doc_.createElement(GSXML.DOCUMENT_ELEM+GSXML.LIST_MODIFIER);
	c.appendChild(document_list);
	// add each document
	for (int d = 0; d < docs.size(); d++) {
	    long docnum = ((MGDocInfo) docs.elementAt(d)).num_;
	    String id = gdbm_src_.docnum2Oid(docnum);
	    System.out.println("Docnum: " + docnum + " ID: " + id);
	    Node no = GSXML.createDocumentElement(doc_, id);
	    document_list.appendChild(no);
	}

	Element term_list = doc_.createElement(GSXML.TERM_ELEM+GSXML.LIST_MODIFIER);
	c.appendChild(term_list);
	// get the terms out, and add to query result metadata
	Vector terms = mqr.getTerms();
	for (int t = 0; t < terms.size(); t++) {
	    String term = ((MGTermInfo) terms.elementAt(t)).term_;
	    long term_freq = ((MGTermInfo) terms.elementAt(t)).term_freq_;
	    int stem_method = ((MGTermInfo) terms.elementAt(t)).stem_method_;

	    Element term_elem = doc_.createElement(GSXML.TERM_ELEM);
	    term_elem.setAttribute(GSXML.NAME_ATT, term);
	    term_elem.setAttribute(COUNT_ATT, "" + term_freq);
	    term_elem.setAttribute(STEM_ATT, "" + stem_method);
	    term_list.appendChild(term_elem);
	}

	return result;
    }


    // should probably use a list rather than map
    protected boolean setStandardQueryParams(HashMap params)
    {
	// set the default ones
	mg_src_.setReturnTerms(true);
	Set entries = params.entrySet();
	Iterator i = entries.iterator();
	while (i.hasNext()) {
	    Map.Entry m = (Map.Entry)i.next();
	    String name = (String)m.getKey();
	    String value = (String)m.getValue();

	    if (name.equals(CASE_PARAM)) {
		boolean val = (value.equals(BOOLEAN_PARAM_ON) ? true : false);
		mg_src_.setCase(val);
	    }
	    else if (name.equals(STEM_PARAM)) {
		boolean val = (value.equals(BOOLEAN_PARAM_ON) ? true : false);
		mg_src_.setStem(val);
	    }
	    else if (name.equals(MATCH_PARAM)) {
		int mode;
		if (value.equals(MATCH_PARAM_ALL)) mode = 1;
		else mode = 0;
		mg_src_.setMatchMode(mode);
	    }
	    else if (name.equals(RANK_PARAM)) {
		if (value.equals(RANK_PARAM_RANK)) {
		    mg_src_.setSortByRank(true);
		} else if (value.equals(RANK_PARAM_NONE)) {
		    mg_src_.setSortByRank(false);
		}
	    }
	    else if (name.equals(MAXDOCS_PARAM)) {
		int docs = Integer.parseInt(value);
		mg_src_.setMaxDocs(docs);
	    } // ignore any others
	}
	return true;
    }
}
